/* OpenCP Module Player
 * copyright (c) '94-'04 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * DGA graphic driver
 *
 * revision history: (please note changes here)
 *  -ss040707   Stian Skjelstad <stian@nixia.no>
 *    -first release
 */

#define _CONSOLE_DRIVER
#include "config.h"
#include <curses.h> /* for key-codes */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#if 0
#include <X11/Xos.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xmu/StdSel.h>
#include <X11/Xmd.h>
#endif
#include <X11/extensions/xf86dga1.h> /* just using the old version query to state a warning for now */
#include <X11/extensions/xf86dga.h>
#if 0
#include <ctype.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if 0
#include <signal.h>
#endif
#include <unistd.h>
#include <sys/types.h>
#include "boot/psetting.h"
#include "types.h"

#include "poutput-fb.h"
#include "boot/console.h"
#include "poutput.h"

#include "pfonts.h"

#define MINMAJOR 2
#define MINMINOR 0

static int plBPP=8;
static int plDepth=8; /* can be 32 if BPP is 24 */
static int plRealBytesPerLine;

static uint16_t red[256]=  {0x0000, 0x0000, 0x0000, 0x0000, 0xaaaa, 0xaaaa, 0xaaaa, 0xaaaa, 0x5555, 0x5555, 0x5555, 0x5555, 0xffff, 0xffff, 0xffff, 0xffff};
static uint16_t green[256]={0x0000, 0x0000, 0xaaaa, 0xaaaa, 0x0000, 0x0000, 0x5555, 0xaaaa, 0x5555, 0x5555, 0xffff, 0xffff, 0x5555, 0x5555, 0xffff, 0xffff};
static uint16_t blue[256]= {0x0000, 0xaaaa, 0x0000, 0xaaaa, 0x0000, 0xaaaa, 0x0000, 0xaaaa, 0x5555, 0xffff, 0x5555, 0xffff, 0x5555, 0xffff, 0x5555, 0xffff};

static uint32_t palette32[256];
static uint16_t palette16[256];
static uint16_t palette15[256];

static Colormap cmap=0;

static Display *mDisplay = NULL;
static int mScreen;
static int mRootWin;
static XDGADevice *mDevice;
static XDGAMode modes[3]; /* lowres = 0, highres = 1 , and the biggest one / default res*/
static int framebufferopen=0;
static uint8_t *virtual_framebuffer, *virtual_framebuffer_end;
static uint8_t currentpage;

static int sloppyreset;
static int brokentrident;

static void __gupdatepal(unsigned char color, unsigned char _red, unsigned char _green, unsigned char _blue)
{
	red[color]=_red<<10;
	green[color]=_green<<10;
	blue[color]=_blue<<10;
}

static void __gflushpal(void)
{
	int i, r, g, b;
	if (plBPP==8)
	{
		cmap=XDGACreateColormap(mDisplay, mScreen, mDevice, AllocAll);
		for (i = 0; i < 256; i++)
		{
			XColor xcol;
			     
			xcol.pixel = i;
			xcol.red = red[i];
			xcol.green = green[i];
			xcol.blue = blue[i];
			xcol.flags = DoBlue | DoGreen | DoRed;
			XStoreColor(mDisplay, cmap, &xcol);
		}
		XDGAInstallColormap(mDisplay, mScreen, cmap); 

		return;
	}
	for (i=0;i<256;i++)
	{
		r=red[i]>>8;
		g=green[i]>>8;
		b=blue[i]>>8;
		palette32[i]=(r<<16)+(g<<8)+b;
		r>>=3;
		g>>=2;
		b>>=3;
		palette16[i]=(r<<11)+(g<<5)+b;
		g>>=1;
		palette15[i]=(r<<10)+(g<<5)+b;
	}
}

static int points(XDGAMode *mode, int X, int Y)
{
	int retval=0;
	if (mode->depth<8)
		return 0;
	if (((mode->viewportWidth>X)&&(mode->viewportHeight>=Y)) || ((mode->viewportWidth>=X)&&(mode->viewportHeight>Y)) )
		retval=2;
	else if ((mode->viewportWidth==X)&&(mode->viewportHeight==Y))
		retval=4;
	if ((mode->depth==8)&&(mode->bitsPerPixel==8))
		retval+=1;
	return retval;
}

static void __gdrawstr(unsigned short y, unsigned short x, const char *str, unsigned short len, unsigned char f, unsigned char b)
{
	unsigned long p=16*y*plScrLineBytes+x*8;
	unsigned char *sp;
	short i,j,k;

	_plSetGraphPage(p>>16);
	sp=(unsigned char*)plVidMem+(p&0xFFFF);

	f=plpalette[f]&0x0f;
	b=plpalette[b]&0x0f;
	for (i=0; i<16; i++)
	{
		const unsigned char *s=(unsigned char *)str;
		for (k=0; k<len; k++)
		{
			unsigned char bitmap=plFont816[*s][i];
			for (j=0; j<8; j++)
			{
				*sp++=(bitmap&128)?f:b;
				bitmap<<=1;
			}
			if (*s)
				s++;
		}
		sp+=plScrLineBytes-8*len;
	}
}

static void __gupdatestr(unsigned short y, unsigned short x, const uint16_t *str, unsigned short len, uint16_t *old)
{
	unsigned long p=16*y*plScrLineBytes+x*8;
	unsigned char *sp;
	short i,j,k;

	_plSetGraphPage(p>>16);
	sp=(unsigned char*)plVidMem+(p&0xFFFF);

	for (k=0; k<len; k++, str++, old++)
		if (*str!=*old)
		{
			unsigned char *bitmap0=plFont816[((unsigned char*)str)[0]];
			unsigned char f=plpalette[((unsigned char*)str)[1]]&0x0F;
			unsigned char b=plpalette[((unsigned char*)str)[1]]>>4;
			*old=*str;

			for (i=0; i<16; i++)
			{
				unsigned char bitmap=bitmap0[i];
				for (j=0; j<8; j++)
				{
					*sp++=(bitmap&128)?f:b;
					bitmap<<=1;
				}
				sp+=plScrLineBytes-8;
			}
			sp-=16*plScrLineBytes-8;
		} else
			sp+=8;
}

static void __gdrawchar8(unsigned short x, unsigned short y, unsigned char c, unsigned char f, unsigned char b)
{
	unsigned char *cp=plFont88[c];
	unsigned long p=y*plScrLineBytes+x;
	unsigned char pg=p>>16;
	unsigned char *scr;
	short i,j;

	_plSetGraphPage(pg);
	scr=(unsigned char*)(plVidMem+(p&0xFFFF));

	f=plpalette[f]&0x0f;
	b=plpalette[b]&0x0f;

	for (i=0; i<8; i++)
	{
		unsigned char bitmap=*cp++;
		for (j=0; j<8; j++)
		{
			if (scr>=(unsigned char*)plVidMem+0x10000)
				scr-=_plSetGraphPage(++pg);
			*scr++=(bitmap&128)?f:b;
			bitmap<<=1;
		}
		scr+=plScrLineBytes-8;
	}
}
static void __gdrawchar8t(unsigned short x, unsigned short y, unsigned char c, unsigned char f)
{
	unsigned char *cp=plFont88[c];
	unsigned long p=y*plScrLineBytes+x;
	unsigned char pg=p>>16;
	char *scr;
	short i,j;

	_plSetGraphPage(pg);
	scr=(char*)(plVidMem+(p&0xFFFF));

	f=plpalette[f]&0x0f;

	for (i=0; i<8; i++)
	{
		unsigned char bitmap=*cp++;
		for (j=0; j<8; j++)
		{
			if (scr>=(char*)plVidMem+0x10000)
				scr-=_plSetGraphPage(++pg);
			if (bitmap&128)
				*scr=f;
			scr++;
			bitmap<<=1;
		}
		scr+=plScrLineBytes-8;
	}
}

static void __gdrawchar8p(unsigned short x, unsigned short y, unsigned char c, unsigned char f, void *picp)
{
	unsigned char *cp=plFont88[c];
	unsigned long p=y*plScrLineBytes+x;
	unsigned char pg=p>>16;
	char *scr;
	char *pic;
	short i,j;

      	if (!picp)
	{
		_gdrawchar8(x,y,c,f,0);
		return;
	}

	f=plpalette[f]&0x0f;
	_plSetGraphPage(pg);
	scr=(char*)(plVidMem+(p&0xFFFF));
	pic=(char*)picp+p;

	for (i=0; i<8; i++)
	{
		unsigned char bitmap=*cp++;
		for (j=0; j<8; j++)
		{
			if (scr>=((char*)plVidMem+0x10000))
				scr-=_plSetGraphPage(++pg);
			if (bitmap&128)
				*scr=f;
			else
				*scr=*pic;
			scr++;
			pic++;
			bitmap<<=1;
		}
		scr+=plScrLineBytes-8;
		pic+=plScrLineBytes-8;
	}
}

static void __gdrawcharp(unsigned short x, unsigned short y, unsigned char c, unsigned char f, void *picp)
{

	unsigned char *cp=plFont816[c];
	unsigned long p=y*plScrLineBytes+x;
	char *pic=(char*)picp+p;
	unsigned char pg=p>>16;
	char *scr;
	short i,j;

      	if (!picp)
	{
		_gdrawchar(x,y,c,f,0);
		return;
	}

	_plSetGraphPage(pg);
	scr=(char*)(plVidMem+(p&0xFFFF));

	f=plpalette[f]&0x0f;

	for (i=0; i<16; i++)
	{
		unsigned char bitmap=*cp++;
		for (j=0; j<8; j++)
		{
			if (scr>=(char*)plVidMem+0x10000)
			{
				scr-=_plSetGraphPage(++pg);
			}
			if (bitmap&128)
				*scr=f;
			else
				*scr=*pic;
			scr++;
			pic++;
			bitmap<<=1;
		}
		pic+=plScrLineBytes-8;
		scr+=plScrLineBytes-8;
	}
}

static void __gdrawchar(unsigned short x, unsigned short y, unsigned char c, unsigned char f, unsigned char b)
{
	unsigned char *cp=plFont816[c];
	unsigned long p=y*plScrLineBytes+x;
	unsigned char pg=p>>16;
	char *scr;
	short i,j;

	f=plpalette[f]&0x0f;
	b=plpalette[b]&0x0f;
	_plSetGraphPage(pg);
	scr=(char*)(plVidMem+(p&0xFFFF));

	for (i=0; i<16; i++)
	{
		unsigned char bitmap=*cp++;
		for (j=0; j<8; j++)
		{
			if (scr>=(char*)(plVidMem+0x10000))
			{
				scr-=_plSetGraphPage(++pg);
			}
			*scr++=(bitmap&128)?f:b;
			bitmap<<=1;
		}
		scr+=plScrLineBytes-8;
	}
}

static int __plSetGraphMode(int high)
{
#ifdef DGA_DEBUG
	if (high==-1)
		fprintf(stderr, "dga: set normale mode\n");
	else if (high)
		fprintf(stderr, "dga: set 1024x768\n");
	else
		fprintf(stderr, "dga: set 640x480\n");
#endif

	if (virtual_framebuffer)
	{
		free(virtual_framebuffer);
		virtual_framebuffer=0;
	}
	
	if (high==-1)
	{
		plVidMem=0;
		if (framebufferopen)
		{
			XUngrabKeyboard(mDisplay, CurrentTime);
/*			XDGACloseFramebuffer(mDisplay, mScreen);*/
			if (sloppyreset)
				XDGASetMode(mDisplay, mScreen, modes[2].num);
			XDGASetMode(mDisplay, mScreen, 0);
			framebufferopen=0;
		}
		return 0;
	}

	if (!framebufferopen)
	{		
/*		if (!XDGAOpenFramebuffer(mDisplay, mScreen))
		{
			fprintf(stderr, "dga: XDGAOpenFramebuffer failed\n");
			return -1;
		}*/
		XGrabKeyboard(mDisplay, mRootWin, True, GrabModeAsync, GrabModeAsync, CurrentTime);
		framebufferopen=1;
	}	

	if (high)
	{
		if (brokentrident)
			XDGASetMode(mDisplay, mScreen, modes[2].num);
		if (!(mDevice=XDGASetMode(mDisplay, mScreen, modes[1].num)))
		{
			fprintf(stderr, "dga: XDGASetMode failed\n");
/*			XDGACloseFramebuffer(mDisplay, mScreen);*/
			framebufferopen=0;
			return -1;
		}
		plScrMode=101;
		plScrWidth=128;
		plScrHeight=60;
		plScrLineBytes=modes[1].pixmapWidth;
		plBPP=modes[1].bitsPerPixel;
		if ((plBPP!=8) || (plScrLineBytes!=1024))
		{
			plRealBytesPerLine=plScrLineBytes;
			plScrLineBytes=1024;
			virtual_framebuffer=calloc(1024*768|65535, 1);
			virtual_framebuffer_end=virtual_framebuffer+1024*768;
		} else
			virtual_framebuffer=0;
		plDepth=modes[1].depth;
	} else {
#ifdef DGA_DEBUG
		fprintf(stderr, "About to set mode %d\n", modes[0].num);
#endif
		if (!(mDevice=XDGASetMode(mDisplay, mScreen, modes[0].num)))
		{
			fprintf(stderr, "dga: XDGASetMode failed\n");
/*			XDGACloseFramebuffer(mDisplay, mScreen);*/
			framebufferopen=0;
			return -1;
		}
		plScrMode=100;
		plScrWidth=80;
		plScrHeight=60;
		plScrLineBytes=modes[0].pixmapWidth;
		plBPP=modes[0].bitsPerPixel;
		if ((plBPP!=8) || (plScrLineBytes!=640))
		{
			plRealBytesPerLine=plScrLineBytes;
			plScrLineBytes=640;
			virtual_framebuffer=calloc(640*480|65535, 1);
			virtual_framebuffer_end=virtual_framebuffer+640*480;
		} else
			virtual_framebuffer=0;
		plDepth=modes[0].depth;
	}

	memset(mDevice->data, 0, plRealBytesPerLine * plScrHeight * 8);
	if (!virtual_framebuffer) /* we have both perfect pitch and 8 bpp */
		plVidMem=(char *)mDevice->data;
	else
		plVidMem=(char *)virtual_framebuffer;
	currentpage=0;

	_gflushpal();
	return 0;
}

static int __plSetGraphPage(unsigned char x)
{
	int retval=0;/*65536;*/
	{
		if (virtual_framebuffer)
		{
			if (x==currentpage)
			{
			} else if (plBPP==32)
			{ 
				int bytestodo=65536;
					
				uint8_t *src=(uint8_t *)plVidMem;

				int startx = (currentpage * 0x10000) % plScrLineBytes;
				int starty = (currentpage * 0x10000) / plScrLineBytes;

				uint32_t *dst_line = ((uint32_t *)mDevice->data) + starty * plRealBytesPerLine;
				uint32_t *dst = dst_line + startx ;

				int j = startx;

				goto tight_32;

				while (1)
				{
					for (j=0;j<plScrLineBytes;j++)
					{
						tight_32:
						*dst=palette32[*src];
						dst++;
						src++;
						if (!--bytestodo)
							goto done_32;
						if (src==virtual_framebuffer_end)
							goto done_32;
					}
					dst = dst_line += plRealBytesPerLine;
				}
				
				done_32:
				plVidMem=(char *)virtual_framebuffer + x * 0x10000;
			} else if (plDepth==16)
			{ 
				int bytestodo=65536;
					
				uint8_t *src=(uint8_t *)plVidMem;

				int startx = (currentpage * 0x10000) % plScrLineBytes;
				int starty = (currentpage * 0x10000) / plScrLineBytes;

				uint16_t *dst_line = ((uint16_t *)mDevice->data) + starty * plRealBytesPerLine;
				uint16_t *dst = dst_line + startx ;

				int j = startx;

				goto tight_16;

				while (1)
				{
					for (j=0;j<plScrLineBytes;j++)
					{
						tight_16:
						*dst=palette16[*src];
						dst++;
						src++;
						if (!--bytestodo)
							goto done_16;
						if (src==virtual_framebuffer_end)
							goto done_16;
					}
					dst = dst_line += plRealBytesPerLine;
				}
				
				done_16:
				plVidMem=(char *)virtual_framebuffer + x * 0x10000;
			} else if (plDepth==15)
			{ 
				int bytestodo=65536;
					
				uint8_t *src=(uint8_t *)plVidMem;

				int startx = (currentpage * 0x10000) % plScrLineBytes;
				int starty = (currentpage * 0x10000) / plScrLineBytes;

				uint16_t *dst_line = ((uint16_t *)mDevice->data) + starty * plRealBytesPerLine;
				uint16_t *dst = dst_line + startx ;

				int j = startx;

				goto tight_15;

				while (1)
				{
					for (j=0;j<plScrLineBytes;j++)
					{
						tight_15:
						*dst=palette15[*src];
						dst++;
						src++;
						if (!--bytestodo)
							goto done_15;
						if (src==virtual_framebuffer_end)
							goto done_15;
					}
					dst = dst_line += plRealBytesPerLine;
				}
				done_15:
				plVidMem=(char *)virtual_framebuffer + x * 0x10000;
			} else if (plDepth==8)
			{ 
				int bytestodo=65536;
					
				uint8_t *src=(uint8_t *)plVidMem;

				int startx = (currentpage * 0x10000) % plScrLineBytes;
				int starty = (currentpage * 0x10000) / plScrLineBytes;

				uint8_t *dst_line = ((uint8_t *)mDevice->data) + starty * plRealBytesPerLine;
				uint8_t *dst = dst_line + startx ;

				int j = startx;

				goto tight_8;

				while (1)
				{
					for (j=0;j<plScrLineBytes;j++)
					{
						tight_8:
						*dst=*src;
						dst++;
						src++;
						if (!--bytestodo)
							goto done_8;
						if (src==virtual_framebuffer_end)
							goto done_8;
					}
					dst = dst_line += plRealBytesPerLine;
				}
				done_8:
				plVidMem=(char *)virtual_framebuffer + x * 0x10000;
			}
		} else {
			plVidMem=(char *)mDevice->data + x * 0x10000;
		}

		currentpage=x;
	}
	{
		int n_chars = 0;
		char buf[21];
		static KeySym ks = 0;

		XEvent event;

		if (!XCheckMaskEvent(mDisplay, 0xffffffff, &event))
			return retval;
		switch (event.type)
		{
			case KeyPress:
				n_chars = XLookupString(&event.xkey, buf, 20, &ks, NULL);
				buf[n_chars] = '\0';
#ifdef DGA_DEBUG
				fprintf(stderr,"KeyPress [%d]: %s\n", event.xkey.keycode, buf);
				fprintf(stderr, "n_chars=%d    %d %d %d %d\n", n_chars, buf[0], buf[1], buf[2], buf[3]);
#endif
				if ((n_chars==1)&&(buf[0]>=32)/*&&(buf[0]<=127)*/)
					___push_key(buf[0]);
				else {
					uint16_t key = event.xkey.keycode;
					switch (key)
					{
						case 9: key=KEY_ESC; break;
						case 22: key=KEY_BS; break;
						case 23: key=KEY_TAB; break;
						case 36: key=_KEY_ENTER; break;
						default:
/*						case 37:
						case 50:
						case 64:*/ key=0; break;
						case 67: key=KEY_F1; break;
						case 68: key=KEY_F2; break;
						case 69: key=KEY_F3; break;
						case 70: key=KEY_F4; break;
						case 71: key=KEY_F5; break;
						case 72: key=KEY_F6; break;
						case 73: key=KEY_F7; break;
						case 74: key=KEY_F8; break;
						case 75: key=KEY_F9; break;
						case 76: key=KEY_F10; break;
						case 95: key=KEY_F11; break;
						case 96: key=KEY_F12; break;
						case 97: key=KEY_HOME; break;
						case 98: key=KEY_UP; break;
						case 99: key=KEY_NPAGE; break;
						case 100: key=KEY_LEFT; break;
						case 102: key=KEY_RIGHT; break;
						case 103: key=KEY_END; break;
						case 104: key=KEY_DOWN; break;
						case 105: key=KEY_PPAGE; break;
						case 106: key=KEY_INSERT; break;
						case 107: key=KEY_DELETE; break;
					}
					if (key)
						___push_key(key);
				}
				break;
			case KeyRelease:
				n_chars = XLookupString(&event.xkey, buf, 20, &ks, NULL);
		}
		return retval;
	}
}
int dga_init(int ignore)
{
	int MajorVersion, MinorVersion;
	int EventBase, ErrorBase;
/*   	int i, bpp;*/
/*	char *addr;
	int width, bank, banks, ram;
	XEvent event;
	Colormap cmap = 0;
	Visual *vis;
	int flags;*/

	sloppyreset=cfGetProfileBool("dga", "sloppyreset", 1, 0);
	if  ((brokentrident=cfGetProfileBool("dga", "brokentrident", 0, 0)))
		sloppyreset=1; /* brokentrident depends on sloppyreset */

	if (sloppyreset)
		fprintf(stderr, "dga: sloppy reset enabled in config\n");
	if (brokentrident)
		fprintf(stderr, "dga: broken trident enabled in config\n");

#ifdef XDGANeedRoot
	fprintf(stderr, "dga: checking if we are suid root... ");
	if (geteuid()) {
		fprintf(stderr, "failed\n");
		return -1;
	}
	fprintf(stderr, "okey\n");
#endif

	if ( (mDisplay = XOpenDisplay(NULL)) == NULL )
	{
		fprintf( stderr, "dga: can't connect to X server %s\n", XDisplayName(NULL));
		return -1;
	}
	fprintf(stderr, "dga: X is online\n");

	mScreen=DefaultScreen(mDisplay);
	mRootWin=RootWindow(mDisplay, mScreen);
	
	if (!XDGAQueryVersion(mDisplay, &MajorVersion, &MinorVersion))
	{
		if (!XF86DGAQueryVersion(mDisplay, &MajorVersion, &MinorVersion))
		{
			fprintf(stderr, "dga: Unable to query video extension version\n");
			goto errorout;
		} else {
			fprintf(stderr, "dga: Old DGA extentions found :-( (%d.%d)\n", MajorVersion, MinorVersion);
			goto errorout;
		}
	}

	fprintf(stderr, "dga: Version %d.%d found\n", MajorVersion, MinorVersion);

	if (!XF86DGAQueryExtension(mDisplay, &EventBase, &ErrorBase)) {
		fprintf(stderr, "dga: Unable to query video extension information\n");
		goto errorout;
	}
	
	{
		int lowres=-1, highres=-1;
		int lowresscore=0, highresscore=0;
		int defres=-1;
		int maxx=-1, maxy=-1;
		XDGAMode *mode;
		int index, i;
		if (!(mode=XDGAQueryModes(mDisplay, mScreen, &index)))
		{
			fprintf(stderr, "dga: Unable to Query Modes\n");
			goto errorout;
		}
		for (i=0;i<index;i++)
		{
			int score;
#ifdef DGA_DEBUG
			fprintf(stderr, "dga: mode %d - %s@%fHz " /*%dx%d vs %dx%d vs %dx%d with*/ "bpp=%d depth=%d (%d bytes per line)\n", mode[i].num, mode[i].name, mode[i].verticalRefresh, /*mode[i].imageWidth, mode[i].imageHeight, mode[i].pixmapWidth, mode[i].pixmapHeight, mode[i].viewportWidth, mode[i].viewportHeight,*/ mode[i].bitsPerPixel, mode[i].depth, mode[i].bytesPerScanline);
#endif
			
			score=points(mode+i, 640, 480);
			if (score>lowresscore)
			{
				lowresscore=score;
				lowres=i;
			}
			score=points(mode+i, 1024, 768);
			if (score>highresscore)
			{
				highresscore=score;
				highres=i;
			}
			if (sloppyreset)
				if ((mode[i].viewportWidth>=maxx)&&(mode[i].viewportHeight>=maxy))
				{
					maxx=mode[i].viewportWidth;
					maxy=mode[i].viewportHeight;
					defres=i;
				}
		}
		if (lowres<0)
		{
			fprintf(stderr, "dga: Failed to find a low resolution\n");
			goto errorout;
		}
		if (highres<0)
		{
			fprintf(stderr, "dga: Failed to find a high resolution\n");
			goto errorout;
		}
#ifdef DGA_DEBUG
		fprintf(stderr, "dga: using lowres=%d, highres=%d (%d %d)\n", mode[lowres].num, mode[highres].num, lowresscore, highresscore);
#endif
		memcpy(&modes[0], &mode[lowres], sizeof(XDGAMode));
		memcpy(&modes[1], &mode[highres], sizeof(XDGAMode));
		if (sloppyreset)
			memcpy(&modes[2], &mode[defres], sizeof(XDGAMode));

	}

	if (!XDGAOpenFramebuffer(mDisplay, mScreen))
	{
		fprintf(stderr, "dga: XDGAOpenFramebuffer failed\n");
		return -1;
	}

	_plSetGraphMode=__plSetGraphMode;
	_plSetGraphPage=__plSetGraphPage;
	_gdrawstr=__gdrawstr;
	_gdrawchar8=__gdrawchar8;
	_gdrawchar8p=__gdrawchar8p;
	_gdrawchar8t=__gdrawchar8t;
	_gdrawcharp=__gdrawcharp;
	_gdrawchar=__gdrawchar;
	_gupdatestr=__gupdatestr;
	_gupdatepal=__gupdatepal;
	_gflushpal=__gflushpal;

	plVidType=vidVESA;
	
	return 0;	
	
errorout:
	XCloseDisplay(mDisplay);

	return -1;
}

void dga_done(void)
{
	if (!mDisplay)
		return;
	__plSetGraphMode(-1);
	XDGACloseFramebuffer(mDisplay, mScreen);
	XCloseDisplay(mDisplay);
	mDisplay=NULL;
}
